% This is a preliminary implementation
% of the method described in the paper
% "Implicit QR with compression",
% by P. Boito, Y. Eidelman, L. Gemignani
% and I. Gohberg.
%
% The proposed method is a variant of
% structured implicit QR, where the
% unitary factors V and F are not explicitly
% used; instead, the generators are
% updated directly and then compressed.
%
% This version implements single shift.
%
% Input data: pol (vector of polynomial coefficients).
% Output : ev (vector of computed eigenvalues).
%
% Last update by Paola, April 18, 2011.
% 

function ev=qrwc(pol)

% write pol in column form if necessary
[pol1,pol2]=size(pol);
if pol1<pol2
    pol=pol.';
end

normA=norm(pol);

n=length(pol)-1;
ev=zeros(n,1); % computed eigenvalues will be stored here
threshold=eps;  % threshold for "magic trick"
iterations=zeros(n,1);  % number of iterations needed for each eigenvalue

%*************************************************
% Initialization of F, V, p, q.
%
% F is 3x3x(n-1),
% V is 2x2x(n-1),
% p and q are vectors of length n
%*************************************************

E2=eye(2);
E3=eye(3);
P3=[0 1 0; 1 0 0; 0 0 1];

F=zeros(3,3,n-1);

F(:,:,1:n-3)=P3(:,:,ones(1,n-3)); % initialize F with permuted identity matrices

F(:,:,n-2)=[0 0 1; 1 0 0; 0 1 0];
F(:,:,n-1)=eye(3);

V=E2(:,:,ones(1,n-1));  % initialize V with identity matrices

q=pol(2:n+1);   % initialize p and q with polynomial coefficients
q(n)=q(n)+1;
p=[1;zeros(n-1,1)];

%*************************************************
% Initialization of g, h , b, sigma.
%
% g is nx2,
% h is 2xn,
% b is 2x2xn,
% sigma is a vector of length n-1.
%*************************************************

% Note that generators of index n have smaller size
% (extra entries are set to zero)
h=zeros(2,n);
g=zeros(n,2);
b=zeros(2,2,n-1);
sigma=zeros(n-1,1);
gamma=zeros(n-1,2);

h(:,1:n-2)=F(1:2,1,1:n-2);
b(:,:,1:n-2)=F(1:2,2:3,1:n-2);

h(:,n-1)=F(1:2,1,n-1);
b(:,:,n-1)=[F(1:2,2,n-1), zeros(2,1)];
h(:,n)=[1; 0];

gamma(1,:)=[0 1];
g(1,:)=[1 0];

for k=1:n-2
    M=V(:,:,k+1)*[gamma(k,:), 0; 0, 0, 1]*F(:,:,k);
    sigma(k)=M(1,1);
    g(k+1,:)=M(1,2:3);
    gamma(k+1,:)=M(2,2:3);
end
sigma(n-1)=gamma(n-1,:)*h(:,n-1);
g(n,:)=gamma(n-1,:)*b(:,:,n-1);

% *** CHECK: compute the companion matrix and verify that generators are
% correct.
CM=compan(pol);
CMgen=compute_matrix(g,h,b,sigma,p,q);
norm(CM-CMgen)
% end check

defl_size=n; % this means we are working on a matrix of size defl_size

while defl_size>1   % begin main loop

    % Compute diagonal and subdiagonal entries
    diagonals=g(:,1).*h(1,:).'+g(:,2).*h(2,:).'-p.*q;
    subdiag=sigma(1:defl_size-1)-p(2:defl_size).*q(1:defl_size-1);
    % display(subdiag(defl_size-1))

    % Look for first (from the bottom) small subdiagonal entry
    % (standard criterion)
    m=defl_size-1;
    while (m>0)&&(abs(subdiag(m))>eps*(abs(diagonals(m))+abs(diagonals(m+1))))
    %while (m>0)&&(abs(subdiag(m))>eps*normA)
    m=m-1;
    end

    % Has an eigenvalue been found?
    if m==defl_size-1
        %fprintf('one eigenvalue found \n')
        ev(defl_size)=diagonals(defl_size);
        defl_size=defl_size-1;
        %display(defl_size)
    else
        % we'll have to perform QR then!
        alpha=diagonals(defl_size); % shift
        iterations(defl_size)=iterations(defl_size)+1;
        if iterations(defl_size)>10 % random shift if needed
            alpha=rand+i*rand;
            iterations(defl_size)=0;
            fprintf('Random shift...\n')
        end
        m=m+1;
        % Check for "magic trick" (two small consecutive subdiagonal entries)
        % in the submatrix of indices m:defl_size
        k=defl_size-1; % now assume that the bulge-chasing process starts from k...
        while k>m
            G=planerot([diagonals(k)-alpha;subdiag(k)]);
            y=[subdiag(k-1);0];
            y=G*y;
            if abs(y(2))<threshold
         %       fprintf('Magic trick!\n')
         %       display(k)
                if (k<defl_size-1)&&(k>2)
          %          fprintf('...with small m! \n')
                end
                break
            end
            k=k-1;
        end
        m=k;
        % We are ready now to perform QR on the submatrix of indices
        % m:defl_size
        [g,h,b,sigma,p,q,CM]=bulge_chasing(g,h,b,sigma,p,q,m,defl_size,alpha,n,CM);
    end
end
ev(1)=diagonals(1);